{Game Doctor Send Component 1.0

Thanks to R. Bubba Magillicutty for his helpful ASM Send source, and to Bung for
releasing the full TPC source which was used to finish up the receive code.

This component will send either single or multi-part roms to Game Doctor family copiers.
It has been tested on the SNES GDSF7 units and will also work on Doctor V64.

USAGE: 1.  To send a single game to the doctor, call Send(FileName)
       2.  To send multiple files (like for SNES) call SendFiles(FileList) where
           Filelist is a TStringlist.
       2.  To receive a file from the doctor,call Receieve(FileName)

KNOWN BUGS
       1.  NT support is slow because TinyPort is really, really slow.
       2.  Receive support is not working, it seems to be dropping a byte
           randomly.  Someone can fix this I hope?

NOTE:  If you want your program to run under Windows NT you must get a special shareware port
       driver called TinyPort.  It also includes a Pascal unit TPINFACE.PAS which you will need.
       Install it properly according to the instructions.
       For example, if your system uses LPT1, you would configure TinyPort as:
       TPCONFIG ISA 0 378 4
       Tinyport is slow....real slow.

Please do not distribute this unit with changes unless:
       1.  You have fixed one of the known bugs
       2.  You have Incremented the version number.
       3.  You are damned sure it works!}

unit gdsend;

interface

{Remove the period below to compile the NT/TinyPort version}
{.$DEFINE NTVERSION}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs
  {$IFDEF NTVERSION},tpInface{$ENDIF};

const LPT1            : Word = $378;
      LPT2            : Word = $278;
      PORT_STROBE            = 0; {Timeout for strobes, 0 for infinite wait}
      PORT_TIMEOUT           = 0; {Timeout for portinit, 0 for infinite wait}
      GD_READ_COMMAND        = 'GD6R';
      GD_WRITE_COMMAND       = 'GD6W'; 
      BLOCK_SIZE             = 32768;

type
  EPortTimeout    = class(Exception);
  TParallelPort   = (ppLPT1,ppLPT2);
  TSendname       = String[11];
  TGameDoctorAction=(gdInitPort,gdSendHeader,gdSending,gdReceiving);
  TProgressEvent  = procedure(Action : TGameDoctorAction;FileName : String;FileNumber, Progress : Byte) of object;
  TGameDoctor = class(TComponent)
  private
    {-Field variables for properties-}
    FPort      : TParallelPort;
    {-Field vars for events-}
    FOnProgress: TProgressEvent;
    {-Internal-use variables-}
    PortBase   : Word;
    {$IFDEF NTVERSION}
    TPHandle   : THandle;
    {$ENDIF}
    {-Internal-use functions and procedures-}
    {Property setters}
    procedure SetPort(Value : TParallelPort);
    {ASM For reading/writing the port}
    function  PortIn(IOAddr : WORD) : BYTE;
    procedure PortOut(IOAddr : WORD; Data : BYTE);
    {Util functions for port stuff}
    procedure PortCheckSync;
    procedure PortWaitFor(IOAddr : Word; Data : Byte;Wait : Longint);
    procedure PortWaitForNot(IOAddr : Word; Data : Byte;Wait : Longint);
    procedure PortSendString(aString : String);
    procedure PortSendInteger(aInt : Longint);
    procedure PortRecvBlock(var aBlock : Array of Byte; Size : Integer);
    procedure PortSendBlock(aBlock : Array of Byte; Size : Integer);

    {Game Doctor functions}
    function  FormatFileName(aString : String) : TSendName;
    procedure InitPort;
    procedure ClosePort;
    procedure SendHeader(FileSize : Longint; FileName : TSendName);
  protected
    { Protected declarations }
  public
    constructor Create(aOwner : TComponent); override;
    procedure Send(Filename : TFileName);
    procedure SendFiles(FileList : TStrings);
    procedure Receive(FileName : TFilename);
  published
    property Port      : TParallelPort  read FPort       write SetPort;
    property OnProgress: TProgressEvent read FOnProgress write FOnProgress;
  end;

procedure Register;

implementation

{Property setters}
procedure TGameDoctor.SetPort(Value : TParallelPort);
begin
  FPort := Value;
  case FPort of
    ppLPT1: PortBase := LPT1;
    ppLPT2: PortBase := LPT2;
  end;
end;

{Assembly functions for port input/output}
function TGameDoctor.PortIn(IOAddr : WORD) : BYTE;
begin
  {$IFDEF NTVERSION}
  if ReadByte(TPhandle,IOAddr,Result) <> ERROR_SUCCESS then
    ShowMessage(Format('Tinyport read of %d failed',[IOAddr]));
  {$ELSE}
  asm
    mov dx,IOAddr
    in  al,dx
    mov result,al
  end;
  {$ENDIF}
end;

procedure TGameDoctor.PortOut(IOAddr : WORD; Data : BYTE);
begin
  {$IFDEF NTVERSION}
  if WriteByte(TPHandle,IOAddr,Data) <> ERROR_SUCCESS then
    ShowMessage(Format('Tinyport write of %d to %d failed',[Data,IOAddr]));
  {$ELSE}
  asm
    mov  dx,IOAddr
    mov  al,Data
    out  dx,al
  end;
  {$ENDIF}
end;

procedure TGameDoctor.PortCheckSync;
var Ticks  : Integer;
    InByte : Byte;
begin
  if PORT_TIMEOUT <> 0 then
    Ticks := PORT_TIMEOUT
  else
    Ticks := 1;
  {1.  Loop until base+2 and 3 = 3 or base+3 and 3=0, if timeout exit with error}
  While Ticks > 0 do
  begin
    InByte := PortIn(PortBase+2);
    if (((InByte and 3)=3) or ((InByte and 0)=0)) then
      Break;
    if PORT_TIMEOUT <> 0 then
      Dec(Ticks);
  end;
  if Ticks=0 then
    Raise EPortTimeout.Create('Error waiting for port response');

  {2.  Output 0 to base}
  PortOut(PortBase,0);

  {3.  Loop until base+1 and 0 =0}
  PortWaitForNot(PortBase+1,0,PORT_TIMEOUT);

end;

procedure TGameDoctor.PortWaitFor(IOAddr : Word; Data : Byte;Wait : Longint);
var Ticks  : Integer;
    InByte : Byte;
begin
  if Wait <> 0 then
    Ticks := Wait
  else
    Ticks := 1;
  While Ticks > 0 do
  begin
    InByte := PortIn(IOAddr);
    if (InByte and Data)<>0 then
      Exit;
    {Adjust timeout unless we passed 0 which is no timeout}
    if Wait <> 0 then
      Dec(Ticks);
  end;
  Raise EPortTimeout.Create('Error waiting for port response');
end;

procedure TGameDoctor.PortWaitForNot(IOAddr : Word; Data : Byte;Wait : Longint);
var Ticks  : Integer;
    InByte : Byte;
begin
  if Wait <> 0 then
    Ticks := Wait
  else
    Ticks := 1;
  While Ticks > 0 do
  begin
    InByte := PortIn(IOAddr);
    if (InByte and Data)=0 then
      Exit;
    {Adjust timeout unless we passed 0 which is no timeout}
    if Wait <> 0 then
      Dec(Ticks);
  end;
  Raise EPortTimeout.Create('Error waiting for port response');
end;

procedure TGameDoctor.PortSendString(aString : String);
var Chars : Array[0..255] of char;
    Bytes : Array[0..255] of Byte absolute Chars;
begin
  StrPCopy(Chars,aString);
  PortSendBlock(Bytes,Length(aString));
end;

procedure TGameDoctor.PortSendInteger(aInt : Longint);
var Bytes : Array[0..3] of Byte absolute aInt;
begin
  PortSendBlock(Bytes,4);
end;

procedure TGameDoctor.PortRecvBlock(var aBlock : Array of Byte; Size : Integer);
var x         : Integer;
    LowNibble : Byte;
    HiNibble  : Byte;
begin
  PortOut(PortBase,$80);
  for x := 0 to Size-1 do
  begin
    LowNibble := $0;
    HiNibble  := $0;

    PortWaitFor(PortBase+1,$80,PORT_TIMEOUT);
    LowNibble := Portin(PortBase+1);
    LowNibble := ((LowNibble shr 3) and $0F);

    PortOut(PortBase,$0);
    PortWaitForNot(PortBase+1,$80,PORT_TIMEOUT);
    HiNibble := PortIN(PortBase+1);
    HiNibble := ((HiNibble shl 1) and $F0);

    aBlock[x] := (HiNibble or LowNibble);
    PortOut(PortBase,$80);
  end;
end;

procedure TGameDoctor.PortSendBlock(aBlock : Array of Byte; Size : Integer);
var FirstByte : Boolean;
    x         : Integer;
begin
  FirstByte := not ((PortIn(PortBase+2) and 1) <> 0);
  x         := 0;
  While x < Size do
  begin
    if Firstbyte then
      begin {Firstbyte}
        PortWaitForNot(PortBase+2,2,PORT_STROBE);
        PortOut(PortBase,aBlock[x]);
        PortOut(PortBase+2,5);
        Inc(x);
        FirstByte := false;
      end
    else
      begin {Otherbyte}
        PortWaitFor(PortBase+2,2,PORT_STROBE);
        PortOut(PortBase,aBlock[x]);
        PortOut(PortBase+2,4);
        Inc(x);
        FirstByte := true;
      end;
  end;
end;

function TGameDoctor.FormatFileName(aString : String) : TSendNAme;
var aPos : Integer;
begin
  aString := UpperCase(ExtractFileName(aString));
  aPos    := Pos('.',aString);
  if aPos <> 0 then
    Result := Copy(aString,1,aPos-1) {Copy filename}
  else
    Result := aString;  
  While Length(Result) < 8 do
    Result := Result+' ';
  if aPos <> 0 then
    Result := Result+Copy(aString,aPos+1,12);
  While Length(Result) < 11 do
    Result := Result+' ';
end;

procedure TGameDoctor.InitPort;
begin
  {$IFDEF NTVERSION}
  if OpenTinyPort(TPHandle) <> ERROR_SUCCESS then
    ShowMessage('Error opening TinyPort driver');
  {$ENDIF}

  {First output 0 to base port, then output 4 to base+2}
  PortOut(PortBase,0);
  PortOut(PortBase+2,4);
  PortWaitForNot(PortBase+2,8,PORT_TIMEOUT);

  {Now output 0AAh to base, then output 0 to base+2}
  PortOut(PortBase,$AA);
  PortOut(PortBase+2,0);
  PortWaitFor(PortBase+2,8,PORT_TIMEOUT);

  {Output 4 to base+2}
  PortOut(PortBase+2,4);
  PortWaitForNot(PortBase+2,8,PORT_TIMEOUT);

  {Output 055h to base, then output 0 to base+2}
  PortOut(PortBase,$55);
  PortOut(PortBase+2,0);
  PortWaitFor(PortBase+2,8,PORT_TIMEOUT);

  {Output 4 to base+2}
  PortOut(PortBase+2,4);
  PortWaitForNot(PortBase+2,8,PORT_TIMEOUT);
end;

procedure TGameDoctor.ClosePort;
begin
  {$IFDEF NTVERSION}
  CloseTinyPort(TPhandle);
  {$ENDIF}
end;

procedure TGameDoctor.SendHeader(FileSize : Longint;FileName : TSendName);
begin
  PortSendInteger(FileSize);
  While Length(FileName) < 11 do
    FileName := FileName+' ';
  PortSendString(FileName);
end;

{-PUBLIC METHODS BEGIN HERE-}

constructor TGameDoctor.Create(aOwner : TComponent);
begin
  inherited Create(aOwner);
  FPort         := ppLPT1;
  PortBase      := LPT1;
end;

procedure TGameDoctor.Send(FileName : TFileName);
var FileList : TStringList;
begin
  FileList := TStringList.Create;
  try
    FileList.Add(Filename);
    SendFiles(FileList);
  finally
    FileList.Free;
  end;
end;

procedure TGameDoctor.SendFiles(FileList : TStrings);
var aFile       : TFileStream;
    Buf         : Array[0..32767] of Byte;
    BufSize     : Integer;
    x           : Integer;
    SendName    : TSendName;
begin
  if FileList.Count=0 then
    Exit;
  try
    {Initialize}
    if Assigned(FOnProgress) then
      FOnProgress(gdInitPort,FileList[0],1,0);
    InitPort;

    {Send the command and # of files}
    if Assigned(FOnProgress) then
      FOnProgress(gdSendHeader,FileList[0],1,0);
    PortSendString(GD_READ_COMMAND+Chr(FileList.Count));

    {Loop and send data}
    for x := 0 to FileList.Count-1 do
    begin
      aFile := TFileStream.Create(FileList[x],fmOpenRead);
      try
        SendName := FormatFileName(FileList[x]);
        SendHeader(aFile.Size,SendName);
        While aFile.Position < aFile.Size do
        begin
          BufSize := aFile.Read(Buf,BLOCK_SIZE);
          PortSendBlock(Buf,BufSize);
          if Assigned(FOnProgress) then
            FOnProgress(gdSending,FileList[x],x+1,Trunc((aFile.Position/aFile.Size)*100));
        end;
      finally
        aFile.Free;
      end;
    end;
  finally
    ClosePort;
  end;
end;

procedure TGameDoctor.Receive(FileName : TFileName);
var aFile       : TFileStream;
    Buf         : Array[0..32767] of Byte;
    BufSize     : Integer;
    x           : Integer;
    FileCount   : Byte;
    FileSizeBuf : Array[0..3] of byte;
    FileSize    : Longint absolute FileSizeBuf;
    TotalSize   : Longint;
    RecvName    : Array[0..10] of char;
    RecvNameBuf : Array[0..10] of byte absolute RecvName;
    FName       : TFileName;
begin
  try
    {Setup}
    StrPCopy(RecvNAme,FormatFileName(FileName));
    {1.  Initialize}
    if Assigned(FOnProgress) then
      FOnProgress(gdInitPort,FileName,1,0);
    InitPort;
    {2.  Send header of GD6W}
    if Assigned(FOnProgress) then
      FOnProgress(gdSendHeader,FileName,1,0);
    PortSendString(GD_WRITE_COMMAND);
    {3.  Send formatted filename}
    PortSendString(RecvName);
    {4.  Sync port}
    PortCheckSync;
    {5.  Read first byte (number of files)}
    PortRecvBlock(FileCount,1);
    {6.  For each file read the size (4 bytes) the filename (11 bytes) and the data}
    for x := 1 to FileCount do
    begin
      {Get the file size and name}
      FillChar(Buf,SizeOf(Buf),$0);
      PortRecvBlock(Buf,15);
      Move(Buf,FileSize,4); {Copy filesize out of the buffer}
      TotalSize := FileSize; {Stored for progress meter}
      Move(Buf[4],RecvNameBuf,11); {Copy the filename out of the buffer}
      {Get the bytes and dump them}
      FName := ExtractFilePath(FileName)+Format('%s.%s',[Copy(RecvName,1,8),Copy(RecvName,9,11)]);
      aFile := TFileStream.Create(Fname,fmCreate);
      While FileSize > 0 do
      begin
        FillChar(Buf,SizeOf(Buf),$0);
        if FileSize > 32768 then
          BufSize := 32768
        else
          BufSize := FileSize;
        PortRecvBlock(Buf,BufSize);
        if Assigned(FOnProgress) then
          FOnProgress(gdSending,FName,x,Trunc((aFile.Size/TotalSize)*100));
        aFile.Write(Buf,BufSize);
        Dec(FileSize,BufSize);
      end;
      aFile.Free;
    end;
  finally
    ClosePort;
  end;
end;

procedure Register;
begin
  RegisterComponents('Console', [TGameDoctor]);
end;

end.

